#!/usr/bin/perl
# Take o65 format relocatable 65816 object files and link them into a LoROM
# SNES image file.
# 2004-01-21
# Based on the 1999 (HiROM) version of ldsnes

# Memory map:
# ,,------ Bank
# vv,,,,-- Address
# 000000-6FFFFF SlowROM equivalent of 800000-EFFFFF
# 700000-707FFF Battery backed SRAM - up to 32kbytes (in cart)
#  70xxxx        Not managed by linker, do what thou wilt
# 708000-77FFFF Battery backed SRAM - same as 700000-707FFFF
# 7E0000-7FFFFF System RAM (Low, High, Expanded) 128kbytes (in console)
#  7E0000-7Exxxx Stack - >=256bytes, <=8192bytes, grows down
#  7E8000-7EFFFF Data segments (copied from rom bank C0 at startup)
#  7F0000-7FFFFF BSS segments
# 800000-801FFF RAM - same as 7E0000-7E1FFF - stack pointer points here
# 802000-807FFF I/O registers
# 808000-80FFFF First 32K of ROM - same as C08000-C0FFFF
# C00000-C07FFF ROM - Same as C08000-C0FFFF (different only on HiROM)
# C08000-C0FFFF First 32K of ROM
# C08000-C0FFFF Empty on emulator due to bug, so we actually use 808xxx instead
#  C08000-C0xxxx Unused space padded with 0xFF characters
#  C0xxxx-C0xxxx Text segments
#  C0xxxx-C0FBFF Data segments (copied to RAM at startup)
#  C0FC00-C0FFFF crt0, ldsnes data, cart data, vectors
#   C0FC00-C0FCFF crt0 (boot loader)
#   C0FD00-C0FEFF reserved
#   C0FF00-C0FFBF ldsnes data
#    C0FF00-C0FF07 magic number "\x06\x09ldsnes"
#    C0FF08-C0FF09 initial stack pointer (highest 7Exxxx addr used for stk)
#    C0FF0A-C0FF0B initial pc (the address of main())
#    C0FF0C-C0FF0D initial pbr (bank address of main())
#    C0FF0E-C0FF0F offset of lowest data segment
#    C0FF10        bank of lowest data segment (=C0) (not used)
#    C0FF11-C0FF12 byte count from lowest data seg to end of highest data seg
#    C0FF13-C0FFBF reserved (irq vectors and whatnot?)
#   C0FFC0-C0FFDF embedded cart data (title, rom size, etc)
#   C0FFFC-C0FFFD reset vector - points to 00FC00 =C0FC00
# C18000-C1FFFF Second 32K of ROM
#  C1xxxx-C1xxxx Unused space padded with 0xFF characters
#  C1xxxx-C1FFFF Text segments
# C28000-C2FFFF Third 32K of ROM...

# Allocated regions will never cross a bank boundary.

# zeropage segments are not allowed. The linker will exit with an error if any
# zeropage regions are requested. You should change your program to allocate
# room on the stack and set the D-register to point to the stack area.

# Stack size is calculated by taking the sum of the requested stack sizes from
# all object files, adding 16 bytes and rounding up to the nearest page. If the
# -s option is supplied, the calculated stack size is discarded and the
# argument to the -s option is used as the number of 256-byte pages to allocate
# for the stack. A warning is printed if the -s option specifies less stack
# space than the calculated value.

# Output filename defaults to a.out and is changable with the -o option.

# The default title to encode in the ROM information area is the uppercased
# version of the output filename. It can be overriden with the -t option.
# Titles are normally in uppercase, but a lowercase string given for -t will
# be preserved. Titles are truncated to 20 characters.

# The ROM type byte defaults to 0 (ROM only). To change it, specify the -T
# option followed by a decimal, 0octal or 0xhex number.

# The ROM size byte is automatically (mis)calculated.

# The RAM size byte defaults to 0. It may be changed with the -R option
# followed by a decimal, 0octal or 0xhex number.

# The country code byte defaults to 1 (USA). It may be changed with the -C
# option followed by a decimal, 0octal or 0xhex number.

# The licensee byte defaults to 0 (undefined). It may be changed with the -L
# option followed by a decimal, 0octal or 0xhex number.

# The game version byte defaults to 0. It may be changed with the -V option
# followed by a decimal, 0octal or 0xhex number.

# The checksum and inverse checksum bytes are (mis)calculated automatically.

# The -v option results in verbosity.

# Bugs:
# This documentation should be cleaned up and converted into perldoc.
# The code is shitty.

use Getopt::Std;

$TEMPDIR="/tmp";
defined($ENV{TEMPDIR}) && do {$TEMPDIR=$ENV{TEMPDIR}};
defined($ENV{TMPDIR}) && do {$TEMPDIR=$ENV{TMPDIR}};
$TEMPFILE="$TEMPDIR/ldsnes.$$.S";

$opts{o}='a.out';

getopts('o:s:t:ST:R:C:L:V:v',\%opts);

if(!defined($opts{t})) {
    $opts{t}=uc($opts{o});
}

if(!defined($opts{T})) {
    $opts{T}="0";
}

if(!defined($opts{R})) {
    $opts{R}="0";
}

if(!defined($opts{C})) {
    $opts{C}="1";
}

if(!defined($opts{L})) {
    $opts{L}="0";
}

if(!defined($opts{V})) {
    $opts{V}="0";
}

@infiles=@ARGV;

# read the files
foreach $file (@infiles) {
    open(INFILE,"<$file") || die "open $file: $!, died";
    read(INFILE,$magic,6) || die "read $file: $!, died";
    $currfile={};
    $hopts=[];
    $$currfile{name}=$file;
    $$currfile{hopts}=$hopts;
    # read header
    if($magic ne "\x01\x00o65\x00") {
	die "$file is not an o65 version 0 file";
    }
    $mode=&readword();
    if(!($mode&0x8000)) {
	die "$file is not for the 65816, died";
    }
    if($mode&0x4000) {
	die "$file does not allow bytewise relocation, died";
	# only reason it matters is because the reloc table is formatted diff
    }
    if($mode&0x2000) {
	die "$file is not 16 bit formatted, died";
    }
    if(!($mode&0x1000)) {
	die "$file is not an object file, died";
    }
    # we do 256 byte alignment no matter what the mode byte says
    $$currfile{tbase}=&readword();
    $tlen=&readword();
    $$currfile{dbase}=&readword();
    $dlen=&readword();
    $$currfile{bbase}=&readword();
    $$currfile{blen}=&readword();
    &readword();
    if(&readword()!=0) {
	die "$file requires a zeropage segment, fix your code. died ";
    }
    $$currfile{stacksz}=&readword();
    # read header options
    for(;;) {
	$len=&readbyte();
	last if($len==0);
	$type=&readbyte();
	if($len==1) {
	    die "malformed header option in $file, died";
	} elsif($len==2) {
	    $bytes='';
	} else {
	    read(INFILE,$bytes,$len-2) || die "read $file: $!, died";
	}
	$thishopt={type=>$type,
		   bytes=>$bytes};
	push @$hopts, $thishopt;
    }
    # read text/data segs
    if($tlen==0) {
	$foo='';
    } else {
	read(INFILE,$foo,$tlen) || die "read $file: $!, died";
    }
    $$currfile{text}=$foo;
    if($dlen==0) {
	$foo='';
    } else {
	read(INFILE,$foo,$dlen) || die "read $file: $!, died";
    }
    $$currfile{data}=$foo;
    # read undefined reference list
    $nudrefs=&readword;
    $udrefs=[];
    $$currfile{udrefs}=$udrefs;
    for($i=0;$i<$nudrefs;$i++) {
	$foo='';
	for(;;) {
	    $foo.=getc(INFILE) || die "read $file: $!, died";
	    if($foo=~/\x00$/) {
		chop($foo);
		last;
	    }
	}
	push @$udrefs, $foo;
    }
    # read relocation table
    $$currfile{textreloc}=&rdreloc($$currfile{tbase});
    $$currfile{datareloc}=&rdreloc($$currfile{dbase});
    # read exported globals list
    $numglobals=&readword();
    $globals=[];
    $$currfile{globals}=$globals;
    for($i=0;$i<$numglobals;$i++) {
	$thisglob={symbol=>&readasciz(),
	           seg=>&readbyte(),
	           value=>&readword()};
	push @$globals, $thisglob;
    }
    push @objects, $currfile;
    close INFILE;
}

if(defined($opts{v})) {
    foreach $object (@objects) {
	print "File $$object{name}\n";
	printf("text segment base=0x%x len=0x%x\n",$$object{tbase},
	       length($$object{text}));
	printf("data segment base=0x%x len=0x%x\n",$$object{dbase},
	       length($$object{data}));
	printf("bss segment base=0x%x len=0x%x\n",$$object{bbase},
	       $$object{blen});
	print "Stack size=$$object{stacksz}\n";
	foreach $hopt (@{$$object{hopts}}) {
	    print "hopt type=$$hopt{type} data=$$hopt{bytes}\n";
	}
	foreach $ref (@{$$object{udrefs}}) {
	    print "undefined ref: $ref\n";
	}
	foreach $reloc (@{$$object{textreloc}}) {
	    printf("text reloc off=0x%x type=0x%x segid=0x%x refidx=0x%x lowbytes=0x%x\n",
		   $$reloc{off},$$reloc{type},$$reloc{segid},$$reloc{refidx},
		   $$reloc{lowbytes});
	}
	foreach $reloc (@{$$object{datareloc}}) {
	    printf("data reloc type=0x%x segid=0x%x refidx=0x%x lowbytes=0x%x\n",
		   $$reloc{type},$$reloc{segid},$$reloc{refidx},$$reloc{lowbytes});
	}
	foreach $global (@{$$object{globals}}) {
	    printf("global symbol=%s seg=0x%x value=0x%x\n",
		   $$global{symbol},$$global{seg},$$global{value});
	}
    }
}

# decide how much space to use for the stack and allocate it. set $rb1 to the
# address of the first unused byte in LowRAM.
$stack=0x100;
foreach $object (@objects) {
    $stack+=$$object{stacksz};
}
# ok i lied all but the last page of lowram can be stack
if($stack>0x1f00) {
    die "too much stack space requested";
}
$stack&=0xff00;
if(defined($opts{s})) {
    if($opts{s}*256<$stack) {
	print STDERR "warning: shrinking stack from $stack bytes\n";
    }
    $stack=$opts{s}*256;
}
$rb1=$stack; # not used...
# the last byte of the stack (grows down) will be $stack-1
if(defined($opts{v})) {
    print "Stack will be $stack bytes.\n";
}

# Allocate all data segments
# dataaddr values get the ROM address, not the ultimate RAM address
# $ceiling[x] is the lowest used address @ top of bank x (x=0xc0, 0xc1, ...)
@ceiling=(0x10000) x 0xff;
# $floor[x] is the lowest available address @ bottom of bank x
@floor=(0x8000) x 0xff;
$ceiling[0x80]=0xfc00;
foreach $object (@objects) {
	$ceiling[0x80]-=length($object->{data});
	if($ceiling[0x80]<$floor[0x80]) {
		die "data segment too large, died";
	}
	$object->{dataaddr}=0x800000+$ceiling[0x80];
}

# Allocate all bss segments
# bssaddr values get the ultimate RAM address
$floor[0x7f]=0;
foreach $object (@objects) {
	$ceiling[0x7f]-=$object->{blen};
	if($ceiling[0x7f]<$floor[0x7f]) {
		die "bss segment too large, died";
	}
	$object->{bssaddr}=0x7f0000+$ceiling[0x7f];
}

# Allocate all text segments
foreach $object (@objects) {
	$whoopee=0;
	for($bank=0x80;$bank<0xef;$bank++) {
		if($ceiling[$bank]-length($object->{text})>=$floor[$bank]) {
			$ceiling[$bank]-=length($object->{text});
			$object->{textaddr}=($bank*0x10000)+$ceiling[$bank];
			$whoopee++;
			last;
		}
	}
	if(!$whoopee) {
		die "unable to allocate room for $$object{name} text segment, died";
	}
}

# All the segments have been allocated; now, add globals to the global table
foreach $object (@objects) {
	foreach $global (@{$$object{globals}}) {
		if(defined($globs{$global->{symbol}})) {
			die "$object->{name} redefines $global->{symbol}. original definition was at $globs{$global->{symbol}}. died";
		}
		if($$global{seg}==2) {
		    $globs{$$global{symbol}}=($$global{value}-$$object{tbase})+
			$$object{textaddr};
		} elsif($$global{seg}==3) {
		    $globs{$$global{symbol}}=($$global{value}-$$object{dbase})+
			$$object{dataaddr};
		} elsif($$global{seg}==4) {
		    $globs{$$global{symbol}}=($$global{value}-$$object{bbase})+
			$$object{bssaddr};
		} else {
		    die "$$object{name} declares global $$global{symbol} in invalid segment $$global{seg}, died";
		}
	}
}

if(defined($opts{v})) {
    foreach $object (@objects) {
	printf("$$object{name} text segment at 0x%x, data at 0x%x, bss at 0x%x\n",
	       $$object{textaddr},$$object{dataaddr},$$object{bssaddr});
    }
    foreach $glob (keys %globs) {
	printf("global $glob is at %x\n",$globs{$glob});
    }
}

# now make the requested relocation fixups
foreach $object (@objects) {
    &popbases();
    foreach $reloc (@{$$object{textreloc}}) {
	&doreloc($reloc,\$$object{text});
    }
    foreach $reloc (@{$$object{datareloc}}) {
	&doreloc($reloc,\$$object{data});
    }
}

# actually write the output file
#open(OUTFILE,">$opts{o}") || die "unable to write to $opts{o}";
open(OUTFILE,"+>$opts{o}") || die "unable to write to $opts{o}";
# "pre-pad" the file by filling it with 0xFF's, then overwrite the used parts
# and while we're at it, measure the file size
$filesize=0;
for($bank=0x80;$bank<0xef;$bank++) {
	if($ceiling[$bank]<0x10000) {
		print OUTFILE "\xFF"x0x8000;
		$filesize+=0x8000;
	}
}
# write the ldsnes data
seek OUTFILE,0xc0ff00-0xc08000,0;
print OUTFILE "\x06\x09ldsnes"; # magic [0xc0ff00]
print OUTFILE pack("v",$stack-1); # initial stack pointer [0xc0ff08]
if(!defined($globs{_main})) {
    die "no _main(), died at";
}
print OUTFILE pack("v",$globs{_main}&0xffff); # initial pc [0xc0ff0a]
print OUTFILE chr(($globs{_main}>>16)&0xff)."\x00"; # initial pbr [0xc0ff0c]
print OUTFILE "\x00\x80"; # data segment start (for start copy code) [0xc0ff0d]
print OUTFILE "\xc0"; # data segment bank [0xc0ff0f]
print OUTFILE "\x00\x80"; # data segment size [0xc0ff10]
# write out the actual segments to the file
foreach $object (@objects) {
	if(length($object->{text})!=0) {
		$bank=$object->{textaddr}>>16;
		$offset=$object->{textaddr}&0xffff;
		seek OUTFILE,(($bank-0x80)*0x8000)+($offset-0x8000),0;
		print OUTFILE $object->{text};
	}
	if(length($object->{data})!=0) {
		$bank=$object->{dataaddr}>>16;
		$offset=$object->{dataaddr}&0xffff;
		seek OUTFILE,(($bank-0x80)*0x8000)+($offset-0x8000),0;
		print OUTFILE $object->{data};
	}
}
# write out the crt0
open CODEFILE,">$TEMPFILE" || die "unable to write temporary code file, died";
print CODEFILE <<'NED';
*=$fc00
clc
xce /* disable emulation mode */
rep #$30:.al:.xl /* 16 bit regs */
/* if FastROM then set the cycle speed to 3.58MHz */
lda $80ffd5 /* rom makeup byte */
and #$00f0 /* rom speed nibble */
beq nofastrom
sep #$20:.as /* 8 bit accumulator */
lda #$1 /* fastrom */
sta $80420d /* cycle speed register */
rep #$20:.al /* back to 16 */
nofastrom:
/* copy data segments in bank 0xc0 to system ram in bank 0x7e */
lda $80ff0e /* src offset within bank */
tax
lda $80ff0e /* dest offset within bank */
tay
lda $80ff11 /* byte count to copy */
mvn $807e /* copy it! */
/* set the stack ptr */
lda $80ff08 /* get stored value */
tcs /* and set it */
/* jump to start of code */
lda $80ff0a /* copy stored start pc to zero page */
sta $000000
lda $80ff0c /* copy stored start psr to zero page */
sta $000002
.byt $dc,$00,$00,$00 /* and jump (jml) */
NED
close CODEFILE;
defined($opts{v}) && do {$verbopt='-v'};
open BINCODE,"xa $verbopt -o - $TEMPFILE|" || die "unable to run xa, died";
undef $/;
$booter=<BINCODE>;
close BINCODE;
if(length($booter)==0) {
	die "the booter doesn't look leet, died";
}
seek OUTFILE,0xc0fc00-0xc08000,0;
print OUTFILE $booter;
unlink $TEMPFILE;
# FIXME drop irq code somewhere
# write the rom nf0z
seek OUTFILE,0x7fc0,0;
$title=substr("$opts{t}                    ",0,20)." ";
&outfoo($title,"Title: %s\n");
seek OUTFILE,0x7fd5,0;
if(defined($opts{S})) {
    &outfoo("\x00","SlowROM\n");
} else {
    &outfoo("\x30","FastROM\n");
}
&outfoo(chr(&getnum($opts{T})),"ROM type: %d\n","C");
$romsize=7;
$romsize++ while(1024<<$romsize < $filesize);
&outfoo(chr($romsize),"ROM size: 8192<<%d bits\n","C");
&outfoo(chr(&getnum($opts{R})),"SRAM size byte: %d\n","C");
&outfoo(chr(&getnum($opts{C})),"Country code: %d\n","C");
&outfoo(chr(&getnum($opts{L})),"Licensee byte: %d\n","C");
&outfoo(chr(&getnum($opts{V})),"Version byte: %d\n","C");
#&outfoo("\x00\x00\x00\x00","FIXME random checksum\n");
seek OUTFILE,0x7ffc,0;
print OUTFILE "\x00\xFC"; # reset vector
$checksum=0;
seek OUTFILE,0,0;
for($i=0;$i<$filesize;$i++) {
	$checksum+=getc(OUTFILE);
	$checksum&=0xffff;
}
seek OUTFILE,0x7fdc,0;
&outfoo(chr($checksum&0xff^0xff).chr(($checksum>>8)^0xff),"Inverse checksum: %04x\n","v");
&outfoo(chr($checksum&0xff).chr($checksum>>8),"Checksum: %04x\n","v");
close OUTFILE;

# read two wrong-endian bytes from INFILE and return an integer
sub readword {
    my $foo;

    read(INFILE,$foo,2) || die "read $file: $!, died";
#printf("deep debug: rw 0x%x\n",unpack("v",$foo));
    return unpack("v",$foo);
}

# read a byte from INFILE and return an integer
sub readbyte {
    my $foo;

    read(INFILE,$foo,1) || die "read $file: $!, died";
#printf("deep debug: rb 0x%x\n",ord($foo));
    return ord($foo);
}

# read an asciz string from INFILE
sub readasciz {
    my $foo;

    $foo='';
    for(;;) {
	$foo.=getc(INFILE) || die "read $file: $!, died";
	if($foo=~/\x00$/) {
	    chop($foo);
	    return $foo;
	}
    }
}

# read a relocation table from INFILE. return it as a ref to a list of hash
# refs. takes base address as arg.
sub rdreloc {
    my($tbl, $ent, $foo);
#    my $off=$_[0];
    my $off=-1; # do offset values relative to start of seg

    $tbl=[];
    for(;;) {
	$ent={};
#	read(INFILE,$foo,1) || die "read $file: $!, died";
#printf("about to read reloc, old off=0x%x\n",$off);
#	$foo=ord($foo);
	$foo=&readbyte();
	if($foo==0) {
	    last;
	} elsif($foo==255) {
	    $off+=254;
#read(INFILE,$foo,1); # throw away a character. dont ask me why
	    next;
	} else {
	    $off+=$foo;
	    $$ent{off}=$off;
#	    read(INFILE,$foo,1) || die "read $file: $!, died";
#	    $foo=ord($foo);
	    $foo=&readbyte();
	    $$ent{type}=$foo&0xe0;
	    $$ent{segid}=$foo&0x1f;
	    if($$ent{segid}==0) { # undefined reference
		$$ent{refidx}=&readword();
	    }
	    if($$ent{type}==0x40) { # high byte reloc
		# the high byte is the one we are changing, so we need to know
		# what the low byte is.
#		read(INFILE,$foo,1) || die "read $file: $!, died";
#		$$ent{lowbytes}=ord($foo);
		$$ent{lowbytes}=&readbyte();
	    } elsif($$ent{type}==0xa0) { # segment reloc
		# we are changing the bank byte so we need to know the 2 low
		# bytes
#		$$ent{lowbytes}=&readword();
#$$ent{lowbytes}=&readbyte()<<8;
# the spec fucking LIES and says that the low bytes are stored for a SEG reloc
# entry, they aren't.
	    }
	    push @$tbl, $ent;
	}
    }
    return $tbl;
}

# fill in @origbase and @newbase with the original (to the assembler) and new
# (after relocation) base addresses for each of the segments for $$object
# i.e. mess around with some icky global arrays to avoid needing to structure
# things right
sub popbases {
    $origbase[2]=$$object{tbase};
    $origbase[3]=$$object{dbase};
    $origbase[4]=$$object{bbase};
    $newbase[2]=$$object{textaddr};
#    $newbase[3]=$$object{dataaddr}-0x430000; # get the ram address not rom
    $newbase[3]=$$object{dataaddr}-0x020000; # get the ram address not rom
    $newbase[4]=$$object{bssaddr};
}

# translate an address from the assembler output to a shiny new relocated
# address
sub transaddr {
    my $addr=shift;
    my $reloc=shift;
    my @udrefs=@{$$object{udrefs}}; # icky global $$object fu
    my $foo;

    if($$reloc{segid}==0) {
	if(!defined($foo=$globs{$udrefs[$$reloc{refidx}]})) {
	    die "unable to find symbol $udrefs[$$reloc{refidx}]";
	}
#printf "resolved symbol $udrefs[$$reloc{refidx}] as %x\n",$foo;
	$addr=$foo;
    } elsif($$reloc{segid}>=2 && $$reloc{segid}<=4) {
	$addr-=$origbase[$$reloc{segid}];
	$addr+=$newbase[$$reloc{segid}];
    } else {
	die "invalid relocation segment $$reloc{segid}";
    }

    return $addr;
}

# Perform a relocation fixup.
sub doreloc {
    my $reloc=shift;
    my $segcont=shift;

    if($$reloc{type}==0x80) { # WORD - 2 byte address
	$addr=unpack("v",substr($$segcont,$$reloc{off},2));
	$addr=&transaddr($addr,$reloc);
#	print "$segcont ".length($$segcont)." $$reloc{off}\n";
	substr($$segcont,$$reloc{off},2)=pack("v",$addr&0xffff);
    } elsif($$reloc{type}==0x40) { # HIGH - high byte of an address
	$addr=(ord(substr($$segcont,$$reloc{off},1))<<8)+
	    $$reloc{lowbytes};
	$addr=&transaddr($addr,$reloc);
	substr($$segcont,$$reloc{off},1)=chr(($addr>>8)&0xff);
    } elsif($$reloc{type}==0x20) { # LOW - low byte of an address
	$addr=ord(substr($$segcont,$$reloc{off},1));
	$addr=&transaddr($addr,$reloc);
	substr($$segcont,$$reloc{off},1)=chr($addr&0xff);
    } elsif($$reloc{type}==0xc0) { # SEGADR - 3 byte address
	$addr=unpack("v",substr($$segcont,$$reloc{off},2));
	$addr+=ord(substr($$segcont,$$reloc{off}+2,1))<<16;
	$addr=&transaddr($addr,$reloc);
	substr($$segcont,$$reloc{off},3)=pack("v",$addr&0xffff)
	    .chr(($addr>>16)&0xff);
    } elsif($$reloc{type}==0xa0) { # SEG - high byte of 3 byte address
	$addr=$$reloc{lowbytes};
$addr=$origbase[$$reloc{segid}];
# sigh sigh sigh sigh sigh sigh sigh. i _think_ that is right.
	$addr+=ord(substr($$segcont,$$reloc{off},1))<<16;
	$addr=&transaddr($addr,$reloc);
	substr($$segcont,$$reloc{off},1)=chr(($addr>>16)&0xff);
    } else {
	die "invalid relocation fixup type $$reloc{type} requested by $$object{name}, died";
    }
}

# Interpret a decimal, 0octal or 0xhex number
sub getnum {
    my $val=shift;

    $val = oct($val) if $val =~ /^0/;
    return $val;
}

# Write a datum to OUTFILE. If verbose, use $format as a printf format to also
# print the datum to stdout. If a third argument is given, use it as an unpack
# format string to convert the datum to human-readable format.
sub outfoo {
    my $datum=shift;
    my $format=shift;
    my $upfmt=shift;

    print OUTFILE $datum;
    if(defined($upfmt)) {
	$datum=unpack($upfmt,$datum);
    }
    if(defined($opts{v})) {
	printf($format,$datum);
    }
}
